@@ -1,4 +1,4 @@
-# Copyright 2022 MosaicML Streaming authors
+# Copyright 2023 MosaicML Streaming authors
 # SPDX-License-Identifier: Apache-2.0
 
 """A mid-epoch-resumable streaming pytorch IterableDataset."""
@@ -8,10 +8,12 @@ import os
 from enum import IntEnum
 from multiprocessing.shared_memory import SharedMemory
 from threading import Thread
-from time import sleep
+from time import sleep, time
 from typing import Any, Dict, Iterator, Optional, Tuple
 
 import numpy as np
+import torch
+import torch.distributed as dist
 from filelock import FileLock
 from numpy.typing import NDArray
 from torch.utils.data import IterableDataset
@@ -202,12 +204,38 @@ class StreamingDataset(IterableDataset):
         # Determine and distribute shuffle seed and shm prefix.
         seed_rng = np.random.default_rng(shuffle_seed)
         self.shuffle_seed = int(seed_rng.integers(1 << 60))
-        prefix_int = np.random.randint(1 << 24)
+        prefix_int = int(seed_rng.integers(1 << 24))
         self._prefix = f'{prefix_int:06x}'
 
         # Should be a unique shared directory per each StreamingDataset instantiation to avoid a conflict
         # between a different StreamingDataset instance on a same machine.
-        self._shared_dir = os.path.join(os.path.sep, 'tmp', 'streaming', self._prefix)
+        start_time = time()
+        while True:
+            self._shared_dir = os.path.join(os.path.sep, 'tmp', 'streaming', self._prefix)
+            if os.path.exists(self._shared_dir):
+                prefix_int = int(seed_rng.integers(1 << 24))
+                self._prefix = f'{prefix_int:06x}'
+            else:
+                break
+            elapsed = time() - start_time
+            # Raise an exception if not finding a unique shared directory in 60 secs
+            if elapsed > 60:
+                raise RuntimeError(''.join([
+                    f'Could not find the unique shared directory, bailing out.',
+                    'Please provide a different `shuffle_seed` value.'
+                ]))
+
+            sleep(TICK)
+
+        # Initialize the distributed package and synchronize all the ranks
+        is_dist_pg_initialized = False
+        if dist.is_available() and not dist.is_initialized():
+            is_dist_pg_initialized = True
+            dist.init_process_group(backend='nccl' if torch.cuda.is_available() and
+                                    dist.is_nccl_available() else 'gloo',
+                                    rank=world.rank,
+                                    world_size=world.num_ranks)
+        dist.barrier()
 
         # Create the shared memory-backed worker barrier, without its lock, which is unpickleable.
         worker_barrier_filelock_path = os.path.join(self._shared_dir, 'barrier_filelock')
@@ -236,6 +264,10 @@ class StreamingDataset(IterableDataset):
         self._shard_states = create_shared_memory(name=f'{self._prefix}_shard_states',
                                                   size=len(self.shard_sizes) * np.uint8(0).nbytes)
 
+        # Destroy process group, and de-initialize the distributed package
+        if is_dist_pg_initialized:
+            dist.destroy_process_group()
+
     @property
     def next_epoch(self) -> int:
         """Get property next_epoch.
@@ -262,19 +294,6 @@ class StreamingDataset(IterableDataset):
         """
         return self.index.get_samples_per_device()
 
-    def __getitem__(self, index: int) -> Dict[str, Any]:
-        """Get sample by global index.
-
-        Args:
-            index (int): Sample index.
-
-        Returns:
-            Dict[str, Any]: Column name with sample data.
-        """
-        shard, index_in_shard = self.index.find_sample(index)
-        reader = self.shards[shard]
-        return reader[index_in_shard]
-
     def _set_canonical_num_nodes(self, world: World):
         """Set the canonical numbers of nodes.
 
@@ -532,14 +551,16 @@ class StreamingDataset(IterableDataset):
             self._download_shard_part(raw_info, zip_info, reader.compression)
 
     def _download_or_skip_shard(self, lock: FileLock, shard_states: NDArray[np.uint8],
-                                shard_id: int) -> None:
-        """Download a shard, skipping if in progress by another worker.
+                                shard_id: int, wait_if_downloading: bool) -> None:
+        """Download a shard, waiting or skipping if in progress by another worker.
 
         Args:
             lock (FileLock): The lock protecting ``shard_states``.
             shard_states (NDArray[np.uint8]): The download status of each shard, as an array in
                 shared memory.
             shard_id (int): Shard ID.
+            wait_if_downloading (bool): Whether to wait or skip if the shard is currently being
+                downloaded by someone else.
         """
         # First, the fast path: check the shared memory shard state without taking the lock. The
         # shard states only ever go up, so if we're at the downloaded state, it's downloaded.
@@ -559,6 +580,9 @@ class StreamingDataset(IterableDataset):
             shard_states[shard_id] = _ShardState.DOWNLOADED
         elif state == _ShardState.DOWNLOADING:
             lock.release()
+            if wait_if_downloading:
+                while shard_states[shard_id] != _ShardState.DOWNLOADED:
+                    sleep(TICK)
         elif state == _ShardState.DOWNLOADED:
             lock.release()
         else:
@@ -579,6 +603,35 @@ class StreamingDataset(IterableDataset):
 
         return lock, shard_states
 
+    def __getitem__(self, idx: int) -> Any:
+        """Get sample by global index, blocking to download its shard if not present.
+
+        Args:
+            idx (int): Sample index.
+
+        Returns:
+            Dict[str, Any]: Mapping of column name to column data.
+        """
+        # Locate the shard and sample offset within that shard where the sample lives.
+        shard_idx, idx_in_shard = self.index.find_sample(idx)
+        shard = self.shards[shard_idx]
+
+        try:
+            # Attempt to directly access the sample for performance reasons.
+            sample = shard[idx_in_shard]
+        except:
+            # Get handles to the shared shard states array and its protective file lock.
+            lock, shard_states = self._get_shard_states()
+
+            # Download the shard if not already being downloaded. Block if download in progress.
+            self._download_or_skip_shard(lock, shard_states, shard_idx, True)
+
+            # Finally, access the sample.
+            sample = shard[idx_in_shard]
+
+        # Return the retrieved sample.
+        return sample
+
     def _download_thread(self, state: _PartitionState) -> None:
         """Download the relevant shards in the background while we are being iterated.
 
@@ -620,7 +673,7 @@ class StreamingDataset(IterableDataset):
 
             # Download and decompress the shard for this sample, if not already done.
             shard_id, _ = self.index.find_sample(sample_id)
-            self._download_or_skip_shard(shard_states_lock, shard_states, shard_id)
+            self._download_or_skip_shard(shard_states_lock, shard_states, shard_id, False)
             state.download_index += 1
 
     def _ready_thread(self, state: _PartitionState) -> None:
