@@ -1,4 +1,4 @@
-# Copyright 2022 MosaicML Streaming authors
+# Copyright 2023 MosaicML Streaming authors
 # SPDX-License-Identifier: Apache-2.0
 
 """:class:`MDSWriter` converts a list of samples into binary `.mds` files that can be read as a :class:`MDSReader`."""
@@ -80,7 +80,10 @@ class MDSWriter(JointWriter):
                 size = len(datum)
                 sizes.append(size)
             else:
-                assert size == len(datum)
+                if size != len(datum):
+                    raise KeyError(
+                        f'Unexpected data size; was this data typed with the correct encoding ({encoding})?'
+                    )
             data.append(datum)
         head = np.array(sizes, np.uint32).tobytes()
         body = b''.join(data)
